# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

import re
import threading
from typing import List

from xmnlp.pinyin.pinyin import Pinyin
from xmnlp.config import path as C_PATH


zh = re.compile(r"([\u4E00-\u9FA5]+)", flags=re.UNICODE)
lock = threading.Lock()
model = None


def loader():
    """load model"""
    with lock:
        global model
        if model is None:
            print("(Lazy Load) Loading model...")
            model = Pinyin()
            model.load(C_PATH.pinyin['model']['pinyin'])


def translate(text: str) -> List[str]:
    """translate chinese to pinyin"""
    loader()
    ret = []
    for s in zh.split(text):
        s = s.strip()
        if not s:
            continue
        if zh.match(s):
            ret += model.translate(s)
        else:
            for word in s.split():
                word = word.strip()
                if word:
                    ret.append(word)
    return ret



# -*- coding: utf-8 -*-

import os

from xmnlp.config import path
from xmnlp.utils import load_stopword


# 模型地址配置
MODEL_DIR = os.getenv('XMNLP_MODEL', None)
ALLOW_POS = ['an', 'i', 'j', 'l', 'n', 'nr', 'ns', 'nt', 'nz',
             't', 'v', 'vd', 'vn', 'x', 'nn', 'g']
SYS_STOPWORDS = load_stopword(path.stopword['corpus']['stopword'])



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

import threading
from typing import List

from xmnlp.config import path as C_PATH
from xmnlp.radical.radical import Radical


model = None
lock = threading.Lock()


def loader():
    """load model"""
    with lock:
        global model
        if model is None:
            print("(Lazy Load) Loading model...")
            model = Radical()
            model.load(C_PATH.radical['model']['radical'])


def radical(text: str) -> List[str]:
    """获取偏旁"""

    loader()
    if not text:
        return None
    return [model.radical(ch) for ch in text]



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# refrence:                                  #
#   https://github.com/isnowfy/snownlp       #
# -------------------------------------------#


from math import log


class BM25:
    def __init__(self, docs, k1=1.5, b=0.75):
        self.N = len(docs)
        self.avgdl = sum([len(doc) + 0.0 for doc in docs]) / self.N
        self.docs = docs
        self.tf = []
        self.df = {}
        self.idf = {}
        self.k1 = k1
        self.b = b
        self.build()

    def build(self):
        for doc in self.docs:
            tmp = {}
            for word in doc:
                if word not in tmp:
                    tmp[word] = 0
                tmp[word] += 1
            self.tf.append(tmp)
            for k, v in tmp.items():
                if k not in self.df:
                    self.df[k] = 0
                self.df[k] += 1
        for k, v in self.df.items():
            self.idf[k] = log(self.N - v + 0.5) - log(v + 0.5)

    def sim(self, doc, idx):
        score = 0
        for word in doc:
            if word not in self.tf[idx]:
                continue
            d = len(self.docs[idx])
            score += (self.idf[word] * self.tf[idx][word] * (self.k1 + 1) / (self.tf[idx][word] + self.k1*(1 - self.b + self.b * d / self.avgdl)))  # NOQA
        return score

    def get_sims(self, doc):
        scores = []
        for idx in range(self.N):
            score = self.sim(doc, idx)
            scores.append(score)
        return scores



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#


class Trie:
    def __init__(self):
        self.root = {}

    def add(self, key, val):
        curr = self.root
        for k in key:
            if k not in curr:
                curr[k] = {}
            curr = curr[k]
        curr['val'] = val

    def find(self, sent, start=0):
        curr = self.root
        ret = None
        for pos in range(start, len(sent)):
            if sent[pos] in curr:
                curr = curr[sent[pos]]
            else:
                return ret
            if 'val' in curr:
                ret = (sent[start: pos + 1], curr['val'])
            else:
                lst = list(curr)
                if lst:
                    cands = {}
                    for c in lst:
                        if 'val' not in curr[c]:
                            continue
                        py = curr[c]['val'][0]
                        if py not in cands:
                            cands[py] = 1
                        else:
                            cands[py] += 1
                    cands = sorted(cands.items(), key=lambda x: x[1], reverse=True)
                    if cands:
                        ret = (sent[start: pos + 1], cands[0][0])
        return ret

    def get(self, sent):
        ret = []
        pos = 0
        while pos < len(sent):
            curr = self.root
            if sent[pos] in curr:
                tmp = self.find(sent, pos)
                if tmp:
                    ret.append(tmp[1])
                    pos += len(tmp[0])
                    continue
            pos += 1
        return ret



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

import os
import threading
from typing import Union, List, Tuple, Dict

from xmnlp import config
from xmnlp.checker.checker import CheckerDecoder


lock = threading.Lock()
checker = None


def load_checker(reload: bool = False) -> None:
    with lock:
        global checker
        if checker is None or reload:
            if config.MODEL_DIR is None:
                raise ValueError("Error: 模型地址未设置，请根据文档「安装」 -> 「下载模型」指引下载并配置模型。")

            print('Lazy load checker...')
            checker = CheckerDecoder(
                os.path.join(config.MODEL_DIR, 'checker'))


def spellcheck(text: str,
               suggest: bool = True,
               k: int = 5,
               max_k: int = 200) -> Union[
                   List[Tuple[int, str]],
                   Dict[Tuple[int, str], List[Tuple[str, float]]]]:
    """spell check
    """
    load_checker()
    return checker.predict(text, suggest=suggest, k=k, max_k=max_k)



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

import os
import threading
from typing import Tuple

from xmnlp import config
from xmnlp.sentiment.sentiment_model import SentimentModel


sentiment_model = None
lock = threading.Lock()


def load_sentiment(reload: bool = False) -> None:
    with lock:
        global sentiment_model
        if sentiment_model is None or reload:
            if config.MODEL_DIR is None:
                raise ValueError("Error: 模型地址未设置，请根据文档「安装」 -> 「下载模型」指引下载并配置模型。")

            print('Lazy load sentiment...')
            sentiment_model = SentimentModel(
                os.path.join(config.MODEL_DIR, 'sentiment'))


def sentiment(doc: str) -> Tuple[float, float]:
    """ 情感分类
    Args:
      doc: str
    Return:
      Tuple[float, float], [proba of negative, proba of postive]
    """
    load_sentiment()
    doc = doc.strip()
    return sentiment_model.predict(doc)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

import logging
from xmnlp.config import path as C_PATH
from xmnlp.pinyin.pinyin import Pinyin
from xmnlp.radical.radical import Radical


logger = logging.getLogger('xmnlp')
logger.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)


class Trainer:
    def __init__(self):
        pass


class PinyinTrainer(Trainer):
    @staticmethod
    def pinyin(srcfile, outfile):
        logger.info('start to train pinyin model...')
        py = Pinyin()
        py.train(srcfile)
        py.save(outfile)
        logger.info('Done !')


class RadicalTrainer(Trainer):
    @staticmethod
    def radical(srcfile, outfile):
        logger.info('start to train radical model...')
        ra = Radical()
        ra.train(srcfile)
        ra.save(outfile)
        logger.info('Done !')


class SysTrainer(Trainer):
    @staticmethod
    def all():
        logger.info('start to train all model...')
        # pinyin
        PinyinTrainer.pinyin(C_PATH.pinyin['corpus']['pinyin'], C_PATH.pinyin['model']['pinyin'])

        # radical
        RadicalTrainer.radical(C_PATH.radical['corpus']['radical'],
                               C_PATH.radical['model']['radical'])

        logger.info('All Done !')



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
#--------------------------------------------#

import os
import concurrent.futures as futures
from functools import partial
from typing import (
    List, Tuple, Generator, Optional, Union, Dict
)

from xmnlp import config
from xmnlp import lexical
from xmnlp import summary
from xmnlp import checker as _checker
from xmnlp import sentiment as _sentiment
from xmnlp import pinyin as _pinyin
from xmnlp import radical as _radical
from xmnlp.utils import load_stopword


__author__ = 'Sean Lee <xmlee97@gmail.com>'
__version__ = '0.3.1'


seg = lexical.seg
tag = lexical.tag
ner = lexical.ner
pinyin = _pinyin.translate
radical = _radical.radical
checker = _checker.spellcheck
sentiment = _sentiment.sentiment


def set_model(dirname: str) -> None:
    config.MODEL_DIR = dirname


def set_stopword(fpath: str) -> None:
    """set stopwords from file"""
    config.SYS_STOPWORDS = set(load_stopword(fpath)) | config.SYS_STOPWORDS


def seg_parallel(texts: List[str], n_jobs: int = 2) -> Generator[List[str], None, None]:
    """seg parallel
    Args:
      texts: List[str]
      n_jobs: int, pool size of threads
    Return:
      Generator[List[str]]
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for words in executor.map(seg, texts):
            yield words


def tag_parallel(texts: List[str], n_jobs: int = 2) -> Generator[List[Tuple[str, str]], None, None]:
    """tag parallel
    Args:
      texts: List[str]
      n_jobs: int, pool size of threads
    Return:
      Generator[List[Tuple[str, str]]]
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for ret in executor.map(tag, texts):
            yield ret


def ner_parallel(texts: List[str], n_jobs: int = 2) -> Generator[List[Tuple[str, str, int, int]], None, None]:
    """ner parallel
    Args:
      texts: List[str]
      n_jobs: int, pool size of threads
    Return:
      Generator[List[Tuple[str, str]]]
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for ret in executor.map(ner, texts):
            yield ret


def keyword(text: str,
            k: int = 10,
            stopword: bool = True,
            allowPOS: Optional[List[str]] = None) -> List[Tuple[str, float]]:
    """extract keyword from text
    Args:
      text: str
      k: int, 返回 topk 个关键词
      stopword: bool, 是否设置停用词
      allowPOS: Optional[List[str]], 允许的词性，默认为系统自定义的词性列表
    Return:
      List[Tuple[str, float]]
    """
    if allowPOS is None:
        allowPOS = config.ALLOW_POS
    stopwords = config.SYS_STOPWORDS if stopword else []

    return summary.keyword(text, k=k, stopword=stopwords, allowPOS=allowPOS)


def keyword_parallel(texts: List[str],
                     k: int = 10,
                     stopword: bool = True,
                     allowPOS: Optional[List[str]] = None,
                     n_jobs: int = 2) -> Generator[List[Tuple[str, float]], None, None]:
    """keyword parallel
    Args:
      texts: List[str]
      k: int, 返回 topk 个关键词
      stopword: bool, 是否设置停用词
      allowPOS: Optional[List[str]], 允许的词性, 默认为系统自定义的词性列表
      n_jobs: int, pool size of threads
    Return:
      List[Tuple[str, float]]
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    keyword_func = partial(keyword, k=k, stopword=stopword, allowPOS=allowPOS)
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for ret in executor.map(keyword_func, texts):
            yield ret


def keyphrase(text: str,
              k: int = 10,
              stopword: bool = False) -> List[str]:
    """keyphrase extraction
    Args:
      text: str
      k: int, 返回 topk 个关键句
      stopword: bool, 是否设置停用词
      allowPOS: Optional[List[str]], 允许的词性, 默认为系统自定义的词性列表
    Return:
      List[str]
    """
    stopwords = config.SYS_STOPWORDS if stopword else []
    return summary.keyphrase(text, k=k, stopword=stopwords)


def keyphrase_parallel(texts: List[str],
                       k: int = 10,
                       stopword: bool = False,
                       n_jobs: int = 2) -> Generator[List[str], None, None]:
    """keyphrase parallel
    Args:
      texts: List[str]
      k: int, 返回 topk 个关键句
      stopword: bool, 是否设置停用词
      n_jobs: int, pool size of threads
    Return:
      List[Tuple[str, float]]
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    keyphrase_func = partial(keyphrase, k=k, stopword=stopword)
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for ret in executor.map(keyphrase_func, texts):
            yield ret


def sentiment_parallel(texts: List[str], n_jobs=2) -> Generator[Tuple[float, float], None, None]:
    """sentiment parallel
    Args:
      texts: List[str]
      n_jobs: int, pool size of threads
    Return:
      Generator[Tuple[float, float], None, None]
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for ret in executor.map(sentiment, texts):
            yield ret


def pinyin_parallel(texts: List[str], n_jobs=2) -> Generator[List[str], None, None]:
    """pinyin parallel
    Args:
      texts: List[str]
      n_jobs: int, pool size of threads
    Return:
      Generator[List[str], None, None]
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for ret in executor.map(pinyin, texts):
            yield ret


def radical_parallel(texts: List[str], n_jobs=2) -> Generator[List[str], None, None]:
    """radical parallel
    Args:
      texts: List[str]
      n_jobs: int, pool size of threads
    Return:
      Generator[List[str], None, None]
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for ret in executor.map(radical, texts):
            yield ret


def checker_parallel(texts: List[str],
                     suggest: bool = True,
                     k: int = 5,
                     max_k: int = 200,
                     n_jobs: int = 2) -> Generator[
                        Union[List[Tuple[int, str]],
                              Dict[Tuple[int, str], List[Tuple[str, float]]]
                        ], None, None]:
    """checker parallel
    Args:
      texts: List[str]
      suggest: bool, 是否返回建议词
      k: int, 返回 topk 个建议词
      max_k: int, 拼音相同词最大检索次数
      n_jobs: int, pool size of threads
    Return:
      Generator
    """
    if not isinstance(texts, list):
        raise ValueError("You should pass a list of texts")
    checker_func = partial(checker, suggest=suggest, k=k, max_k=max_k)
    with futures.ThreadPoolExecutor(max_workers=n_jobs) as executor:
        for ret in executor.map(checker_func, texts):
            yield ret



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

from xmnlp.module import Module


class Radical(Module):
    __notsave__ = []
    __onlysave__ = ['dictionary']

    def __init__(self):
        self.dictionary = {}

    def train(self, fpath):
        """train model"""

        for fname in self.filelist(fpath):
            with open(fname, 'r', encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    arr = line.split(',')
                    if len(arr) != 2:
                        continue
                    self.dictionary[arr[0]] = arr[1]

    def radical(self, char):
        """get radical of given char"""

        if char in self.dictionary:
            return self.dictionary[char]
        return None



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#


import numpy as np
from xmnlp.utils.bm25 import BM25


class KeywordTextRank:
    def __init__(self, words, window=5, alpha=0.85, iters=300):
        self.words = words
        self.window = window
        self.d = alpha
        self.vertex = set(words)
        self.edge = {}
        self.iters = iters

        self.build_edge()
        self.build_matrix()
        self.calc_pr()

    def build_edge(self):
        N = len(self.words)
        for idx, word in enumerate(self.words):
            if word not in self.edge.keys():
                tmp = set()
                l = idx - self.window + 1  # left window
                r = idx + self.window     # right window
                l = 0 if l < 0 else l
                r = N if r >= N else r
                # get window
                for i in range(l, r):
                    if i == idx:
                        continue
                    tmp.add(self.words[i])
                self.edge[word] = tmp

    def build_matrix(self):
        self.matrix = np.zeros([len(self.vertex), len(self.vertex)])
        self.word_idx = {}  # 记录词的idx
        self.idx_dict = {}  # 记录节点idx对应的词
        for i, v in enumerate(self.vertex):
            self.word_idx[v] = i
            self.idx_dict[i] = v
        for key in self.edge.keys():
            for w in self.edge[key]:
                self.matrix[self.word_idx[key]][self.word_idx[w]] = 1
                self.matrix[self.word_idx[w]][self.word_idx[key]] = 1
        for j in range(self.matrix.shape[1]):
            Z = 0
            for i in range(self.matrix.shape[0]):
                Z += self.matrix[i][j]
            for i in range(self.matrix.shape[0]):
                self.matrix[i][j] /= Z

    def calc_pr(self):
        self.PR = np.ones([len(self.vertex), 1])
        for _ in range(self.iters):
            self.PR = (1 - self.d) + self.d * np.dot(self.matrix, self.PR)

    def topk(self, k):
        word_pr = {}
        for i in range(len(self.PR)):
            word_pr[self.idx_dict[i]] = self.PR[i][0]
        res = sorted(word_pr.items(), key=lambda x: x[1], reverse=True)
        return res[:k]


class TextRank:
    def __init__(self, docs, alpha=0.85, min_diff=1e-2, iters=500):
        self.docs = docs
        self.bm25 = BM25(docs)
        self.N = len(docs)
        self.d = alpha
        self.weight = []
        self.weight_sum = []
        self.vertex = []
        self.iters = iters
        self.min_diff = min_diff

        self.build()
        self.calc_pr()

    def build(self):
        for idx, doc in enumerate(self.docs):
            scores = self.bm25.get_sims(doc)
            self.weight.append(scores)
            self.weight_sum.append(sum(scores) - scores[idx])
            self.vertex.append(1.0)

    def calc_pr(self):
        for _ in range(self.iters):
            m = []
            max_diff = 0
            for i in range(self.N):
                m.append(1 - self.d)
                for j in range(self.N):
                    if j == i or self.weight_sum[j] == 0:
                        continue
                    m[-1] += (self.d * self.weight[j][i] / self.weight_sum[j] * self.vertex[j])
                if abs(m[-1] - self.vertex[i]) > max_diff:
                    max_diff = abs(m[-1] - self.vertex[i])
            self.vertex = m
            if max_diff <= self.min_diff:
                break

    def topk(self, k):
        res = list(enumerate(self.vertex))
        res = sorted(res, key=lambda x: x, reverse=True)
        return list(map(lambda x: x[0], res))[:k]



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

import os
import re
import threading
from typing import List, Tuple

from xmnlp import config
from xmnlp.lexical.lexical_model import LexicalDecoder


NER_RULES = [
    ("EMAIL", re.compile(r"[a-zA-Z0-9\-_\.]+@[a-zA-Z0-9\-_]+\.[a-zA-Z0-9\-_\.]+", re.IGNORECASE)),
    ("URL", re.compile(r"(?:http|ftp)s?://"  # http:// or https://
                       r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{1,}\.?)|" # domain
                       r"localhost|"  # localhost
                       r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # or ip
                       r"(?::\d+)?"  # optional port
                       r"(?:/?|[/?]\S+)[a-zA-Z0-9/]+", re.IGNORECASE)),
    ("BOOK", re.compile(r"《.+》"))
]
TAG_MAP = {
    'ORGANIZATION': 'nt',
    'TIME': 't',
    'JOB': 'nn',
    'PERSON': 'nr',
    'LOCATION': 'ns'
}
lexical = None
lock = threading.Lock()


def load_lexical(reload: bool = False) -> None:
    with lock:
        global lexical
        if lexical is None or reload:
            if config.MODEL_DIR is None:
                raise ValueError("Error: 模型地址未设置，请根据文档「安装」 -> 「下载模型」指引下载并配置模型。")

            print('Lazy load lexical...')
            lexical = LexicalDecoder(
                os.path.join(config.MODEL_DIR, 'lexical'), starts=[0], ends=[0])


def seg(doc: str) -> List[str]:
    """分词
    """
    load_lexical()
    doc = doc.strip()
    return [w for w, _ in lexical.predict(doc)]


def tag(doc: str) -> List[Tuple[str, str]]:
    """词性标注
    Args:
      doc: str
    Return: List[Tuple[str, str]], e.g., [(word, tag), ...]
    """
    load_lexical()
    return [(w, TAG_MAP.get(t, t)) for w, t in lexical.predict(doc)]


def ner(doc: str) -> List[Tuple[str, str, int, int]]:
    """命名体识别
    支持的实体类型为:
      (ORGANIZATION, 机构), (TIME, 时间), (JOB, 职业), (PERSON, 人名), (LOCATION, 地点),
      (EMAIL, 邮件), (URL, 网址), (BOOK, 书)

    Args:
      doc: str
    Return: List[Tuple[str, str, int, int]], e.g., [(word, tag, start index, end index), ...]
    """
    load_lexical()
    doc = doc.strip()
    ret = []
    # rule
    for (mathches, tag) in [(r[1].finditer(doc), r[0]) for r in NER_RULES]:
            for mathched in mathches:
                span = mathched.span()
                ret.append((doc[span[0]: span[1]], tag, span[0], span[1]))
    # model
    start_position = 0
    for w, t in lexical.predict(doc):
        if t in TAG_MAP:
            ret.append((w, t, start_position, start_position + len(w)))
        start_position += len(w)
    
    ret.sort(key=lambda x: (x[2], x[3]))
    return ret



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

from xmnlp.module import Module
from xmnlp.utils.trie import Trie


class Pinyin(Module):
    __notsave__ = []
    __onlysave__ = ['trie']

    def __init__(self):
        self.trie = Trie()

    def train(self, fpath):
        """train pinyin model"""

        for fname in self.filelist(fpath):
            with open(fname, 'r', encoding='utf-8') as f:
                for line in f:
                    words = line.split()
                    self.trie.add(words[0], words[1:])

    def translate(self, text):
        """translate chinese charactor to pinyin """
        ret = []
        for t in self.trie.get(text):
            if isinstance(t, (list, tuple)):
                ret = ret + t
            else:
                ret.append(t)
        return ret



# -*- coding: utf-8 -*-

""" XMNLP - Sentiment

Model Tree:

sentiment
├── saved_model.pb
├── variables
│   ├── variables.data-00000-of-00001
│   └── variables.index
└── vocab.txt

"""

import os

import numpy as np
try:
    import tensorflow.compat.v1 as tf
except ImportError:
    import tensorflow as tf
from tokenizers import BertWordPieceTokenizer


MAX_LEN = 150


class SentimentModel:
    def __init__(self, model_dir):
        # load session and graph
        self.sess = tf.Session(graph=tf.Graph())
        tf.saved_model.loader.load(self.sess, ['serve'], export_dir=model_dir)
        self.tokenizer = BertWordPieceTokenizer(os.path.join(model_dir, 'vocab.txt'))
        self.tokenizer.enable_truncation(max_length=MAX_LEN)

    def predict(self, text):
        tokenized = self.tokenizer.encode(text)
        token_ids = tokenized.ids
        segment_ids = tokenized.type_ids
        token_ids, segment_ids = np.array([token_ids]), np.array([segment_ids])
        # placeholder
        input_token = self.sess.graph.get_tensor_by_name('Input-Token:0')
        input_segment = self.sess.graph.get_tensor_by_name('Input-Segment:0')
        output = self.sess.graph.get_tensor_by_name('label/Softmax:0')

        probas = self.sess.run([output], feed_dict={input_token: token_ids,
                                                    input_segment: segment_ids})
        return tuple(probas[0][0].tolist())



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

from typing import List, Tuple, Optional

import xmnlp
from xmnlp.summary.textrank import TextRank, KeywordTextRank
from xmnlp.utils import split_text


def keyword(text: str,
            k: int = 10,
            stopword: Optional[List[str]] = None,
            allowPOS: Optional[List[str]] = None) -> List[Tuple[str, float]]:
    """关键词抽取
    Args:
      text: str, 输入文本
      k: int, 返回 topk 个关键词
      stopword: List[str], 关键词列表，默认 None
      allowPos: List[str], 允许的词性
    Return:
      List[Tuple[str, float]], e.g., [('word', weight), ...]
    """
    text = text.strip()
    if stopword is None:
        stopword = []
    if allowPOS is None:
        allowPOS = []
    words = []
    for ret in xmnlp.tag_parallel(split_text(text)):
        for w, t in ret:
            if w not in stopword and t in allowPOS:
                words.append(w)

    words = KeywordTextRank(words)
    return words.topk(k)


def keyphrase(text: str,
              k: int = 10,
              stopword: Optional[List[str]] = None) -> List[str]:
    """关键句抽取
    Args:
      text: str, 输入文本
      k: int, 返回 topk 个关键词
      stopword: List[str], 关键词列表，默认 None
    Return:
      List[str]
    """
    text = text.strip()
    if stopword is None:
        stopword = []

    docs = []
    for sent in split_text(text):
        words = []
        for word, _ in xmnlp.tag(sent):
            if word not in stopword:
                words.append(word)
        docs.append(words)

    tr = TextRank(docs)
    res = []
    for idx in tr.topk(k):
        res.append(''.join(docs[idx]))

    return res



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

import os
import re
from typing import List, Generator

import numpy as np


re_line_skip = re.compile('[\r\n]')
re_delimiter = re.compile('[，。？！；]')


def split_text(doc: str) -> List[str]:
    sents = []
    for line in re_line_skip.split(doc):
        line = line.strip()
        if not line:
            continue
        for sent in re_delimiter.split(line):
            sent = sent.strip()
            if not sent:
                continue
            sents.append(sent)
    return sents


def filelist(path: str) -> Generator[str, None, None]:
    if os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            if not dirs:
                for f in files:
                    yield os.sep.join([root, f])
    else:
        yield path


def load_stopword(fpath: str) -> List[str]:
    """load stopwords from file """
    stopwords = set()
    for fname in filelist(fpath):
        with open(fname, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                stopwords.add(line)
    return stopwords


def rematch(offsets):
    """ rematch bert token
    """
    mapping = []
    for offset in offsets:
        if offset[0] == 0 and offset[1] == 0:
            mapping.append([])
        else:
            mapping.append([i for i in range(offset[0], offset[1])])
    return mapping


def topK(matrix, K, axis=1):
    """ numpy topK
    """
    if axis == 0:
        row_index = np.arange(matrix.shape[1 - axis])
        topk_index = np.argpartition(-matrix, K, axis=axis)[0:K, :]
        topk_data = matrix[topk_index, row_index]
        topk_index_sort = np.argsort(-topk_data,axis=axis)
        topk_data_sort = topk_data[topk_index_sort,row_index]
        topk_index_sort = topk_index[0:K,:][topk_index_sort,row_index]
    else:
        column_index = np.arange(matrix.shape[1 - axis])[:, None]
        topk_index = np.argpartition(-matrix, K, axis=axis)[:, 0:K]
        topk_data = matrix[column_index, topk_index]
        topk_index_sort = np.argsort(-topk_data, axis=axis)
        topk_data_sort = topk_data[column_index, topk_index_sort]
        topk_index_sort = topk_index[:,0:K][column_index,topk_index_sort]
    return topk_data_sort, topk_index_sort



# -*- coding: utf-8 -*-

""" XMNLP - Spell Checker

Model Tree:

checker
├── corrector
│   ├── saved_model.pb
│   └── variables
│       ├── variables.data-00000-of-00001
│       └── variables.index
├── detector
│   ├── saved_model.pb
│   └── variables
│       ├── variables.data-00000-of-00001
│       └── variables.index
└── vocab.tx

"""

import os

import numpy as np
try:
    import tensorflow.compat.v1 as tf
except ImportError:
    import tensorflow as tf
from tokenizers import BertWordPieceTokenizer

import xmnlp
from xmnlp.utils import rematch, topK


MAX_LEN = 512


class DetectorModel:
    def __init__(self, model_dir):
        # load session and graph
        self.sess = tf.Session(graph=tf.Graph())
        tf.saved_model.loader.load(self.sess, ['serve'], export_dir=model_dir)

    def predict(self, token_ids, segment_ids):
        # placeholder
        input_token = self.sess.graph.get_tensor_by_name('Input-Token:0')
        input_segment = self.sess.graph.get_tensor_by_name('Input-Segment:0')
        output = self.sess.graph.get_tensor_by_name('labels/Sigmoid:0')

        return self.sess.run([output], feed_dict={input_token: token_ids,
                                                  input_segment: segment_ids})


class CorrectorModel:
    def __init__(self, model_dir):
        # load session and graph
        self.sess = tf.Session(graph=tf.Graph())
        tf.saved_model.loader.load(self.sess, ['serve'], export_dir=model_dir)

    def predict(self, token_ids, segment_ids):
        # placeholder
        input_token = self.sess.graph.get_tensor_by_name('Input-Token:0')
        input_segment = self.sess.graph.get_tensor_by_name('Input-Segment:0')
        output = self.sess.graph.get_tensor_by_name('MLM-Activation/truediv:0')

        return self.sess.run([output], feed_dict={input_token: token_ids,
                                                  input_segment: segment_ids})


class CheckerDecoder:
    def __init__(self, model_dir):
        self.detector = DetectorModel(os.path.join(model_dir, 'detector'))
        self.corrector = CorrectorModel(os.path.join(model_dir, 'corrector'))
        self.tokenizer = BertWordPieceTokenizer(os.path.join(model_dir, 'vocab.txt'))
        mask_id = self.tokenizer.encode('[MASK]').ids[1:-1]
        assert len(mask_id) == 1
        self.mask_id = mask_id[0]

    def predict(self, text, suggest=False, k=5, max_k=200):
        tokenized = self.tokenizer.encode(text)
        if len(tokenized.tokens) > MAX_LEN:
            raise ValueError('The text is too long (>512) to process')
        token_ids = tokenized.ids
        segment_ids = tokenized.type_ids
        mapping = rematch(tokenized.offsets)
        token_ids, segment_ids = np.array([token_ids]), np.array([segment_ids])
        probas = self.detector.predict(token_ids, segment_ids)[0][0]
        incorrect_ids = np.where(probas > 0.5)[0]
        token_ids[0, incorrect_ids] = self.mask_id

        if not suggest:
            ret = []
            for i in incorrect_ids:
                ret.append((i - 1, tokenized.tokens[i]))
            return ret

        probas = self.corrector.predict(token_ids, segment_ids)[0][0]
        sorted_probas, sort_indexs = topK(probas, max_k)
        ret = {}
        for i in incorrect_ids:
            if i == 0 or i == len(tokenized.tokens) - 1:
                continue
            current_token = text[mapping[i][0]: mapping[i][-1] + 1]
            current_pinyin = ' '.join(xmnlp.pinyin(current_token))
            cands = []
            for proba, token in zip(sorted_probas[i], self.tokenizer.decode(sort_indexs[i]).split()):
                pinyin = ' '.join(xmnlp.pinyin(token))
                score = 0
                if current_pinyin == pinyin:
                    score = 1
                cands.append((token, proba + score))
            cands.sort(key=lambda x: x[1], reverse=True)
            ret[(i - 1, current_token)] = cands[:k]
        return dict(ret)



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

import os
import bz2
import pickle


class Module:
    __notsave__ = []
    __onlysave__ = []

    def filelist(self, fpath):
        """get file list from filename"""
        if os.path.isdir(fpath):
            for root, dirs, files in os.walk(fpath):
                if not dirs:
                    for f in files:
                        yield os.sep.join([root, f])
        else:
            yield fpath

    def save(self, fname, iszip=True):
        """save model"""
        d = {}
        for k, v in self.__dict__.items():
            if self.__onlysave__:
                if k not in self.__onlysave__:
                    continue
            elif k in self.__notsave__:
                continue

            if isinstance(v, set):
                d[k] = list(v)
            elif hasattr(v, '__dict__'):
                d[k] = v.__dict__
            else:
                d[k] = v

        if not iszip:
            with open(fname, 'wb') as wf:
                pickle.dump(d, wf, True)
        else:
            f = bz2.BZ2File(fname, 'wb')
            f.write(pickle.dumps(d))
            f.close()

    def load(self, fname, iszip=True):
        """load model"""
        if not iszip:
            d = pickle.load(open(fname, 'rb'))
        else:
            try:
                f = bz2.BZ2File(fname, 'rb')
                d = pickle.loads(f.read())
            except IOError:
                f = open(fname, 'rb')
                d = pickle.loads(f.read())
            f.close()
        for k, v in d.items():
            if isinstance(self.__dict__[k], set):
                self.__dict__[k] = set(v)
            elif hasattr(self.__dict__[k], '__dict__'):
                self.__dict__[k].__dict__ = v
            else:
                self.__dict__[k] = v



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#
import os

# mode dir
ROOT = os.sep.join(os.path.dirname(os.path.abspath(__file__)).split(os.sep)[:-1])
CORPUS = os.path.join(ROOT, 'corpus')
CHECKER = os.path.join(ROOT, 'checker')
SENTIMENT = os.path.join(ROOT, 'sentiment')
PINYIN = os.path.join(ROOT, 'pinyin')
RADICAL = os.path.join(ROOT, 'radical')


pinyin = {
    'corpus': {
        'pinyin': os.path.join(CORPUS, 'pinyin'),
    },
    'model': {
        'pinyin': os.path.join(PINYIN, 'pinyin.pickle')
    }
}

checker = {
    'corpus': {
        'checker': os.path.join(CHECKER, 'words.txt'),
    },
}


radical = {
    'corpus': {
        'radical': os.path.join(CORPUS, 'radical'),
    },
    'model': {
        'radical': os.path.join(RADICAL, 'radical.pickle')
    }
}

stopword = {
    'corpus': {
        'stopword': os.path.join(ROOT, 'stopword.txt'),
    }
}



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

""" XMNLP - Lexical

Model Tree:

lexical
├── label2id.json
├── saved_model.pb
├── trans.npy
├── variables
│   ├── variables.data-00000-of-00001
│   └── variables.index
└── vocab.txt
"""

import os
import json

import numpy as np
try:
    import tensorflow.compat.v1 as tf
except ImportError:
    import tensorflow as tf
from tokenizers import BertWordPieceTokenizer

from xmnlp.utils import rematch


MAX_LEN = 512


class LexicalModel:
    def __init__(self, model_dir):
        # load session and graph
        self.sess = tf.Session(graph=tf.Graph())
        tf.saved_model.loader.load(self.sess, ['serve'], export_dir=model_dir)

    def predict(self, token_ids, segment_ids):
        # placeholder
        input_token = self.sess.graph.get_tensor_by_name('Input-Token:0')
        input_segment = self.sess.graph.get_tensor_by_name('Input-Segment:0')
        output = self.sess.graph.get_tensor_by_name('crf/sub_1:0')

        return self.sess.run([output], feed_dict={input_token: token_ids,
                                                  input_segment: segment_ids})


class LexicalDecoder:
    def __init__(self, model_dir, starts=None, ends=None):
        self.trans = np.load(os.path.join(model_dir, 'trans.npy'))
        self.tokenizer = BertWordPieceTokenizer(os.path.join(model_dir, 'vocab.txt'))
        self.lexical_model = LexicalModel(os.path.join(model_dir))
        with open(os.path.join(model_dir, 'label2id.json'), encoding='utf-8') as reader:
            label2id = json.load(reader)
            self.id2label = {int(v): k for k, v in label2id.items()}
        self.num_labels = len(self.trans)
        self.non_starts = []
        self.non_ends = []
        if starts is not None:
            for i in range(self.num_labels):
                if i not in starts:
                    self.non_starts.append(i)
        if ends is not None:
            for i in range(self.num_labels):
                if i not in ends:
                    self.non_ends.append(i)

    def decode(self, nodes):
        """An elegant viterbi decode implementation

        Modified from https://github.com/bojone/bert4keras/blob/master/bert4keras/snippets.py#L404
        """
        # 预处理
        nodes[0, self.non_starts] -= np.inf
        nodes[-1, self.non_ends] -= np.inf

        # 动态规划
        labels = np.arange(self.num_labels).reshape((1, -1))
        scores = nodes[0].reshape((-1, 1))
        paths = labels
        for i in range(1, len(nodes)):
            M = scores + self.trans + nodes[i].reshape((1, -1))
            idxs = M.argmax(0)
            scores = M.max(0).reshape((-1, 1))
            paths = np.concatenate([paths[:, idxs], labels], 0)

        # 最优路径
        return paths[:, scores[:, 0].argmax()]

    def predict(self, text):
        tokenized = self.tokenizer.encode(text)
        if len(tokenized.tokens) > MAX_LEN:
            raise ValueError('The text is too long (>512) to process')
        token_ids = tokenized.ids
        segment_ids = tokenized.type_ids
        mapping = rematch(tokenized.offsets)
        token_ids, segment_ids = np.array([token_ids]), np.array([segment_ids])
        nodes = self.lexical_model.predict(token_ids, segment_ids)[0][0]
        labels = self.decode(nodes)
        entities, starting = [], False
        for i, label in enumerate(labels):
            if label > 0:
                if label % 2 == 1:
                    starting = True
                    entities.append([[i], self.id2label[(label - 1) // 2]])
                elif starting:
                    entities[-1][0].append(i)
                else:
                    starting = False
            else:
                starting = False

        return [(text[mapping[w[0]][0]:mapping[w[-1]][-1] + 1], l)
                for w, l in entities]
